"""doc"""
from .qr_code import qr_terminal
from .type_check import default_str
from .date_util import timestamp_to_date
# from .type_check import type_check

__all__ = [
    'qr_terminal',
    'default_str',
    'timestamp_to_date'
    # 'type_check'
]
